/* KallistiOS 1.1.6

   g2.c
   (c)2000-2001 Dan Potter
*/

/* This module contains low-level G2 bus handling. Right now this is just for
   G2 interrupts, but it will eventually include G2 DMA as well. */

/* Small interrupt listing (from the two Marcus's =)
   For now, everything is mapped thru IRQ 9

  691x -> irq 13
  692x -> irq 11
  693x -> irq 9

  69x0
    bit 2	render complete
        3	scanline 1
        4	scanline 2
        5	vsync
        7	opaque polys accepted
        8	opaque modifiers accepted
        9	translucent polys accepted
        10	translucent modifiers accepted
        12	maple dma complete
        13	maple error(?)
        14	gd-rom dma complete
        15	aica dma complete
        16	external dma 1 complete
        17	external dma 2 complete
        18	??? dma complete
        21	punch-thru polys accepted

  69x4
    bit 0	gd-rom command status
        1	AICA
        2       modem?
        3       expansion port (PCI bridge)

  69x8
    bit 2	out of primitive memory
    bit 3	out of matrix memory

 */

#include <string.h>
#include <stdio.h>
#include <arch/irq.h>
#include <dc/g2.h>
#include <arch/spinlock.h>

CVSID("$Id: g2.c,v 1.3 2002/01/06 01:14:47 bardtx Exp $");

/* Exception table -- this table matches each potential G2 event to a function
   pointer. If the pointer is null, then nothing happens. Otherwise, the
   function will handle the exception. */
static g2evt_handler g2evt_handlers[0x4][0x20];

/* Set a handler, or remove a handler */
int g2evt_set_handler(uint32 code, g2evt_handler hnd) {
	uint32 evtreg, evt;
	
	evtreg = (code >> 8) & 0xff;
	evt = code & 0xff;
	
	if (evtreg > 0x4)
		return -1;
	if (evt > 0x20)
		return -1;

	g2evt_handlers[evtreg][evt] = hnd;
	
	return 0;
}

/* The G2 event handler; this is called from the global IRQ handler
   to handle external IRQ 9. */
static void handler_irq9(irq_t source, irq_context_t *context) {
	int	reg, i;
	vuint32	*asicint = (vuint32*)0xa05f6900;
	uint32	mask;

	/* Go through each event register and look for pending events */
	for (reg=0; reg<4; reg+=2) {
		/* Read the event mask and clear pending */
		mask = asicint[4*reg]; asicint[4*reg] = mask;
		
		/* Search for relevant handlers */
		for (i=0; i<32; i++) {
			if (mask & (1 << i)) {
				if (g2evt_handlers[reg][i] != NULL) {
					g2evt_handlers[reg][i]( (reg << 8) | i );
				}
			}
		}
	}
}


/* Disable all G2 events */
void g2evt_disable_all() {
	vuint32	*asicint = (vuint32*)0xa05f6900;
	
	asicint[0x30/4] = 0;
	asicint[0x30/4] = 0;
	asicint[0x30/4] = 0;
	asicint[0x30/4] = 0;
}

/* Disable a particular G2 event */
void g2evt_disable(uint32 code) {
	vuint32 *asicint = (vuint32*)0xa05f6900;
	asicint[0x30/4 + ((code >> 8) & 0xff)] &= ~(1 << (code & 0xff));
}

/* Enable a particular G2 event */
void g2evt_enable(uint32 code) {
	vuint32 *asicint = (vuint32*)0xa05f6900;
	asicint[0x30/4 + ((code >> 8) & 0xff)] |= (1 << (code & 0xff));
}

/* Initialize events */
void g2evt_init() {
	vuint32 *asic;

	asic = (vuint32*)0xa05f0000;

	/* Clear any pending interrupts and disable all events */
	asic[0x6884/4] = 0;
	asic[0x6910/4] = 0;
	asic[0x6914/4] = 0;
	asic[0x6918/4] = 0;
	asic[0x6920/4] = 0;
	asic[0x6924/4] = 0;
	asic[0x6928/4] = 0;
	asic[0x6930/4] = 0;
	asic[0x6934/4] = 0;
	asic[0x6938/4] = 0;
	asic[0x6900/4] = 0xffffffff;
	asic[0x6904/4] = 0xffffffff;
	asic[0x6908/4] = 0xffffffff;
	
	/* Clear out the event table */
	memset(g2evt_handlers, 0, sizeof(g2evt_handlers));
	
	/* Hook IRQ9 */
	irq_set_handler(EXC_IRQ9, handler_irq9);
}

/* Shutdown events */
void g2evt_shutdown() {
	/* Disable all events */
	g2evt_disable_all();
	
	/* Unhook handler */
	irq_set_handler(EXC_IRQ9, NULL);
	
}

static spinlock_t dma_lock;

/* Lock G2 DMA */
void g2_dma_lock() {
	spinlock_lock(&dma_lock);
}

/* Unlock G2 DMA */
void g2_dma_unlock() {
	spinlock_unlock(&dma_lock);
}

/* Init routine */
void g2_init() {
	g2evt_init();
	spinlock_init(&dma_lock);
}


void g2_shutdown() {
	g2evt_shutdown();
}





